from typing import Any, Union, Dict, List, Callable
import warnings

from collections import namedtuple

from .api import API, FileToken
from .resource import Resource, _ResourceList
from .tag import Tag


class PostNote:
    Point = namedtuple("PostNotePoint", ["x", "y"])

    def __init__(self, polygon: List[List[float]], text: str):
        self._points = [PostNote.Point(*x) for x in polygon]
        self._text = text

    @property
    def points(self) -> List:  # -> List[PostNote.Point]
        return self._points

    @property
    def text(self) -> str:
        return self._text

    @property
    def json(self):
        return {"polygon": [list(p) for p in self._points], "text": self._text}


class Post(Resource):
    @staticmethod
    def _validate_safety(safety: str) -> None:
        if safety not in ("safe", "sketchy", "unsafe"):
            raise ValueError("Safety must be of value safe, sketchy, or unsafe")

    def _str_to_tag(self, val: str) -> Tag:
        t = Tag(self._api, {"names": [val]})
        t.pull()
        return t

    # Implementing Abstract Methods
    def _get_instance_urlparts(self) -> List[str]:
        return ["post", str(self._json["id"])]

    @classmethod
    def _get_class_urlparts(cls) -> List[str]:
        return ["posts"]

    @classmethod
    def _lazy_load_components(cls) -> List[str]:
        return ["id", "safety", "type", "contentUrl", "flags", "tags", "relations"]

    def _setter_transforms(self) -> Dict[str, Callable]:
        return {
            "tags": lambda x: {"names": x.names, "category": x.category},
            "relations": lambda x: {"id": x.id_},
        }

    def _getter_transforms(self) -> Dict[str, Callable]:
        return {
            "tags": lambda x: Tag(self._api, x),
            "relations": lambda x: Post(self._api, x),
        }

    def _serialized(self) -> Dict[str, Any]:
        ret = self._copy_new_json(
            [
                "tags",
                "safety",
                "source",
                "relations",
                "flags",
                "contentToken",
                "thumbnailToken",
                "notes",
            ]
        )
        if "tags" in ret:
            ret["tags"] = [tag["names"][0] for tag in ret["tags"]]
        if "relations" in ret:
            ret["relations"] = [post["id"] for post in ret["relations"]]
        return ret

    # Factory Methods
    @classmethod
    def from_id(cls, api: API, id_: int):  # -> Post
        warnings.warn(
            "Post.from_id() is deprecated, use API.getPost() instead", DeprecationWarning
        )
        p = cls(api, {"id": id_})
        p.pull()
        return p

    @classmethod
    def new(cls, api: API, content: FileToken, safety: str):  # -> Post
        warnings.warn(
            "Post.new() is deprecated, use API.createPost() instead", DeprecationWarning
        )
        cls._validate_safety(safety)
        p = cls(api, {})
        p._json_new = {
            "tags": [],
            "safety": safety,
            "contentToken": content.token,
        }
        p.push()
        return p

    # Getters and Setters
    @property
    def id_(self) -> int:
        return self._generic_getter("id")

    @id_.setter
    def id_(self, val: int) -> None:
        self._generic_setter("id", val)

    @property
    def safety(self) -> str:
        return self._generic_getter("safety")

    @safety.setter
    def safety(self, val: str) -> None:
        self._validate_safety(val)
        self._generic_setter("safety", val)

    @property
    def source(self) -> List[str]:
        return _ResourceList(
            getter=lambda: (self._generic_getter("source") or "").splitlines(),
            parent_resource=self,
            property_name="source",
        )

    @source.setter
    def source(self, val: List[str]) -> None:
        self._generic_setter("source", "\n".join(val))

    @property
    def tags(self) -> List[Tag]:
        return _ResourceList(
            getter=lambda: self._generic_getter("tags"),
            parent_resource=self,
            property_name="tags",
        )

    @tags.setter
    def tags(self, val: List[Union[Tag, str]]) -> None:
        self._generic_setter(
            "tags", [self._str_to_tag(x) if isinstance(x, str) else x for x in val]
        )

    @property
    def relations(self) -> List:  # -> List[Post]
        return _ResourceList(
            getter=lambda: self._generic_getter("relations"),
            parent_resource=self,
            property_name="relations",
        )

    @relations.setter
    def relations(self, val: List) -> None:  # val: List[Post]
        self._generic_setter("relations", val)

    @property
    def content(self) -> str:
        return self._file_getter("content")

    @content.setter
    def content(self, val: FileToken):
        self._file_setter("content", val)

    @property
    def thumbnail(self) -> str:
        return self._file_getter("thumbnail")

    @thumbnail.setter
    def thumbnail(self, val: FileToken):
        self._file_setter("thumbnail", val)

    @property
    def type_(self) -> str:
        return self._generic_getter("type")

    @property
    def mime(self) -> str:
        return self._generic_getter("mimeType")

    @property
    def checksum(self):
        return self._generic_getter("checksum")

    @property
    def width(self) -> int:
        return self._generic_getter("canvasWidth")

    @property
    def height(self) -> int:
        return self._generic_getter("canvasHeight")

    # Flag Getters and Setters
    def _flag_getter(self, flag_name: str) -> bool:
        flag_list = self._generic_getter("flags")
        return flag_name in flag_list

    def _flag_setter(self, flag_name: str, val: bool) -> None:
        flag_list = self._generic_getter("flags")
        if val:
            if flag_name not in flag_list:
                flag_list.append(flag_name)
        else:
            if flag_name in flag_list:
                flag_list.remove(flag_name)
        self._generic_setter("flags", flag_list, False)

    @property
    def loop(self) -> bool:
        return self._flag_getter("loop")

    @loop.setter
    def loop(self, val: bool) -> None:
        self._flag_setter("loop", val)

    @property
    def notes(self) -> List[PostNote]:
        return _ResourceList(
            getter=lambda: [PostNote(**note) for note in self._generic_getter("notes")],
            parent_resource=self,
            property_name="notes",
        )

    @notes.setter
    def notes(self, val: List[PostNote]):
        self._generic_setter("notes", [x.json for x in val])

    @property
    def sound(self) -> bool:
        return self._flag_getter("sound")

    @sound.setter
    def sound(self, val: bool) -> None:
        self._flag_setter("sound", val)

    def __str__(self) -> str:
        return f"Post {self.id_}"

    def __repr__(self) -> str:
        return f"{self._api!r}.<Post of id {self.id_}>"



from typing import List, Generator
import warnings

from collections import namedtuple

from tqdm import tqdm

from .api import API, FileToken
from .resource import Resource
from .tag import Tag
from .post import Post


SearchResult = namedtuple("SearchResult", ["post", "distance", "exact"])


# https://stackoverflow.com/a/45187287/13027787
class _NullContextManager(object):
    def __init__(self, dummy_resource=None):
        self.dummy_resource = dummy_resource

    def __enter__(self):
        return self.dummy_resource

    def __exit__(self, *args):
        pass


def _search_generic(
    api: API,
    search_query: str,
    transforming_class: type,
    page_size: int,
    show_progress_bar: bool = False,
    eager_load: bool = False,
) -> Generator[Resource, None, None]:
    offset = 0
    total = None
    with tqdm() if show_progress_bar else _NullContextManager() as pbar:
        while True:
            urlquery = {"offset": offset, "limit": page_size, "query": search_query}
            if not eager_load:
                urlquery["fields"] = ",".join(transforming_class._lazy_load_components())
            page = api._call(
                "GET",
                transforming_class._get_class_urlparts(),
                urlquery=urlquery,
            )
            offset = offset + len(page["results"])
            if page["total"] != total:
                total = page["total"]
                if show_progress_bar:
                    pbar.total = total
                    pbar.refresh()
            for item in page["results"]:
                if show_progress_bar:
                    pbar.update()
                yield transforming_class(api, item)
            if offset >= total:
                break


def search_tag(
    api: API, search_query: str, page_size: int = 20, show_progress_bar: bool = False
) -> Generator[Tag, None, None]:
    warnings.warn(
        "search_tag() is deprecated, use API.search_tag() instead", DeprecationWarning
    )
    return _search_generic(api, search_query, Tag, page_size, show_progress_bar)


def search_post(
    api: API, search_query: str, page_size: int = 20, show_progress_bar: bool = False
) -> Generator[Post, None, None]:
    warnings.warn(
        "search_post() is deprecated, use API.search_post() instead", DeprecationWarning
    )
    return _search_generic(api, search_query, Post, page_size, show_progress_bar)


def search_by_image(api: API, image: FileToken) -> List[SearchResult]:
    warnings.warn(
        "search_by_image() is deprecated, use API.search_by_image() instead",
        DeprecationWarning,
    )
    result = api._call(
        "POST", ["posts", "reverse-search"], body={"contentToken": image.token}
    )
    ret = [
        SearchResult(post=Post(api, x["post"]), distance=x["distance"], exact=False)
        for x in result["similarPosts"]
    ]
    if result["exactPost"]:
        ret.insert(
            0, SearchResult(post=Post(api, result["exactPost"]), distance=None, exact=True)
        )
    return ret



from typing import Any, Union, Dict, List, BinaryIO

from base64 import b64encode
import json
import os
import re
import urllib.parse

from appdirs import user_data_dir
import requests


class FileToken:
    def __init__(self, token: str, filepath: str):
        self._token = token
        self._filepath = filepath

    @property
    def token(self):
        return self._token

    @property
    def filepath(self):
        return self._filepath

    def __str__(self) -> str:
        return f"<Upload token for file at {self._filepath}>"


class SzurubooruHTTPError(requests.exceptions.HTTPError):
    pass


class API:
    _token_checker = re.compile(
        r"^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$"
    )

    @staticmethod
    def _encode_auth_headers(u: str, p: str) -> str:
        return b64encode(f"{u}:{p}".encode("utf-8")).decode("ascii")

    @staticmethod
    def _check_api_response(r: requests.models.Response) -> None:
        if not r.status_code == requests.codes.ok:
            try:
                msg = r.json()
                msg = f"{msg['name']}: {msg['description']}"
            except ValueError:
                msg = r.text
            raise SzurubooruHTTPError(msg)

    def __init__(
        self,
        base_url: str,
        username: str = None,
        password: str = None,
        token: str = None,
        api_url: str = "api",
    ):
        # Extract Base URL parts
        parsed_base_url = urllib.parse.urlsplit(base_url)
        self._url_scheme = parsed_base_url.scheme
        self._url_netloc = parsed_base_url.netloc
        if not (self._url_scheme and self._url_netloc):
            raise ValueError("Base URL is not valid")
        if self._url_scheme not in ("http", "https"):
            raise ValueError("Base URL must be of HTTP or HTTPS scheme")
        self._url_path_prefix = parsed_base_url.path.rstrip("/")

        # Extract API URL parts
        parsed_api_url = urllib.parse.urlsplit(api_url)
        self._api_scheme = parsed_api_url.scheme or self._url_scheme
        self._api_netloc = parsed_api_url.netloc or self._url_netloc
        if self._api_scheme not in ("http", "https"):
            raise ValueError("API URL must be of HTTP or HTTPS scheme")
        if parsed_api_url.path.startswith("/") or self._api_netloc != self._url_netloc:
            self._api_path_prefix = parsed_api_url.path.rstrip("/")
        else:
            self._api_path_prefix = f"{self._url_path_prefix}/{parsed_api_url.path}".rstrip(
                "/"
            )

        # Extract Auth Info
        self._api_headers = {"Accept": "application/json"}
        self.username = (
            username or parsed_api_url.username or parsed_base_url.username or None
        )
        password = password or parsed_api_url.password or parsed_base_url.password
        if token:
            if not self.username:
                raise ValueError("Token authentication specified without username")
            if not self._token_checker.match(token):
                raise ValueError("Malformed Token String")
            self._api_headers[
                "Authorization"
            ] = f"Token {self._encode_auth_headers(self.username, token)}"
        elif password:
            if not self.username:
                raise ValueError("Password authentication specified without username")
            self._api_headers[
                "Authorization"
            ] = f"Basic {self._encode_auth_headers(self.username, password)}"
        elif self.username:
            raise ValueError("Username specified without authentication method")

    def _create_api_url(self, parts: List[str], query: Dict[str, str] = None) -> str:
        path = [self._api_path_prefix] + [
            urllib.parse.quote(str(part), safe="") for part in parts
        ]
        if query:
            path.append("")
            query = urllib.parse.urlencode(query)
        return urllib.parse.urlunsplit(
            (self._api_scheme, self._api_netloc, "/".join(path), query, None)
        )

    def _call(
        self,
        method: str,
        urlparts: List[str],
        urlquery: Dict[str, str] = None,
        body: Dict[str, Any] = None,
    ) -> Dict[str, Any]:
        req_kwargs = {"headers": self._api_headers}
        if body:
            req_kwargs["json"] = body
        response = requests.request(
            method, self._create_api_url(urlparts, urlquery), **req_kwargs
        )
        self._check_api_response(response)
        return response.json()

    def upload_file(self, file: Union[BinaryIO, str]) -> FileToken:
        if isinstance(file, str):
            with open(file, "rb") as f:
                return self.upload_file(f)
        response = requests.post(
            self._create_api_url(["uploads"]),
            files={"content": file},
            headers=self._api_headers,
        )
        self._check_api_response(response)
        return FileToken(
            response.json()["token"], file.name if hasattr(file, "name") else None
        )

    def _create_data_url(self, rel_url: str, override_base: bool = True) -> str:
        if override_base:
            base_path = os.path.join("/", self._url_path_prefix)
            rel_path = urllib.parse.urlsplit(rel_url).path
            return urllib.parse.urlunsplit(
                (
                    self._url_scheme,
                    self._url_netloc,
                    os.path.join(base_path, rel_path),
                    None,
                    None,
                )
            )
        else:
            return urllib.parse.urljoin(
                urllib.parse.urlunsplit(
                    (self._url_scheme, self._url_netloc, self._url_path_prefix, None, None)
                ),
                rel_url,
            )

    @classmethod
    def save_to_config(cls, config_name: str, **constructor_args) -> None:
        if not config_name.isalnum():
            raise ValueError("config_name must be alphanumeric")
        pathdir = user_data_dir(appname="pyszuru", appauthor=False, roaming=False)
        os.makedirs(pathdir, exist_ok=True)
        path = os.path.join(pathdir, f"{config_name}.json")
        with open(path, "w") as f:
            json.dump(constructor_args, f)

    @classmethod
    def load_from_config(cls, config_name: str):  # -> API:
        if not config_name.isalnum():
            raise ValueError("config_name must be alphanumeric")
        pathdir = user_data_dir(appname="pyszuru", appauthor=False, roaming=False)
        path = os.path.join(pathdir, f"{config_name}.json")
        with open(path, "r") as f:
            constructor_args = json.load(f)
        return cls(**constructor_args)

    def __str__(self) -> str:
        return f"Szurubooru API for {self.username} at {self._api_netloc}"



from typing import List, Generator

from .api import API as _API, FileToken, SzurubooruHTTPError
from .resource import Resource, ResourceNotSynchronized
from .post import Post, PostNote
from .tag import Tag
from .search import search_tag, search_post, search_by_image, _search_generic, SearchResult


class API(_API):
    def getPost(self, id_: int) -> Post:
        p = Post(self, {"id": id_})
        p.pull()
        return p

    def createPost(self, content: FileToken, safety: str) -> Post:
        Post._validate_safety(safety)
        p = Post(self, {})
        p._json_new = {
            "tags": [],
            "safety": safety,
            "contentToken": content.token,
        }
        p.push()
        return p

    def getTag(self, id_: str) -> Tag:
        t = Tag(self, {"names": [id_]})
        t.pull()
        return t

    def createTag(self, name: str) -> Tag:
        # Get default tag category
        tag_cats = [
            x["name"]
            for x in self._call("GET", ["tag-categories"])["results"]
            if x["default"]
        ]
        assert len(tag_cats) == 1
        default_cat = tag_cats[0]
        # Create and return tag
        t = Tag(self, {})
        t._json_new = {"names": [name], "category": default_cat}
        t.push()
        return t

    def search_tag(  # noqa: F811
        self,
        search_query: str,
        page_size: int = 20,
        show_progress_bar: bool = False,
        eager_load: bool = False,
    ) -> Generator[Tag, None, None]:
        return _search_generic(
            self, search_query, Tag, page_size, show_progress_bar, eager_load
        )

    def search_post(  # noqa: F811
        self,
        search_query: str,
        page_size: int = 20,
        show_progress_bar: bool = False,
        eager_load: bool = False,
    ) -> Generator[Post, None, None]:
        return _search_generic(
            self, search_query, Post, page_size, show_progress_bar, eager_load
        )

    def search_by_image(self, image: FileToken) -> List[SearchResult]:  # noqa: F811
        result = self._call(
            "POST", ["posts", "reverse-search"], body={"contentToken": image.token}
        )
        ret = [
            SearchResult(post=Post(self, x["post"]), distance=x["distance"], exact=False)
            for x in result["similarPosts"]
        ]
        if result["exactPost"]:
            ret.insert(
                0,
                SearchResult(
                    post=Post(self, result["exactPost"]), distance=None, exact=True
                ),
            )
        return ret
